"""
VencoPy.

Vehicle Energy Consumption in Python. Electric vehicle fleet load shifting simulations.
"""

__version__ = "0.1.0"
__author__ = 'Niklas Wulff'
__credits__ = 'German Aerospace Center (DLR)'
